#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 9 14:20:17 2019

@author: Sven Serneels, Ponalytics
"""

__name__ = "dicomo"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "1.0.3"
__date__ = "2021-04-07"






