#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 22 12:17:17 2018

@author: Sven Serneels, Ponalytics
"""

__name__ = "plot"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "0.9.0"
__date__ = "2020-04-18"




