#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 12 2020

This folder is temporary. It copies a fix to ipopt: 
    https://github.com/matthias-k/optpy/blob/master/optpy/jacobian.py
    
Folder will stay in direpack until the latter has been released.

"""

__name__ = "opt_temp"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "0.0.2"
__date__ = "2021-04-15"


