# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 17:22:09 2020

@author: Emmanuel Jordy Menvouta
"""



__name__ = "utils"
__author__ = "Emmanuel Jordy Menvouta and Sven Serneels"
__license__ = "MIT"
__version__ = "0.0.2"
__date__ = "2020-12-24"
